#include <stdint.h>
#include <stdlib.h>
#include <string.h>

#include "sin.h"
#include "light.h"

#include "animation.h"

uint16_t timers[4] = { 0, 0, 0, 0 };

enum animation_type animation_type;
enum animation_type previous_animation_type;

// Initialise the animations.
void animation_init(void) {
	// Reset the previous animation type to something invalid.
	previous_animation_type = -1;
	// Set the initial animation type.
	animation_type = 0;
}

void animate(enum animation_event event);

// Run the animations for one tick.
void animation_tick(void) {
	// If the animation type has changed, fire a reset event.
	if (animation_type != previous_animation_type) {
		// Clear all timers.
		memset(timers, 0, sizeof(timers));
		// Send the "reset" event.
		animate(ANIMATION_EVENT_RESET);
		// Remember the animation type.
		previous_animation_type = animation_type;
	} else {
		// Handle each timer in turn.
		for (uint8_t timer = 0; timer < 4; ++timer) {
			if (timers[timer] > 0) {
				if (--timers[timer] == 0) {
					animate(ANIMATION_EVENT_TIMER_A + timer);
				}
			}
		}
	}
}

// Advance to the next animation.
void animation_next_type(void) {
	animation_type = (animation_type + 1) % ANIMATION_COUNT;
}

// Switch to a random animation.
void animation_random_type(void) {
	previous_animation_type = animation_type;
	do {
		animation_type = rand() % ANIMATION_COUNT;
	} while (previous_animation_type == animation_type);
}

// Sets a timer.
void animate_set_timer(enum animation_timer timer, uint16_t ticks) {
	timers[timer] = ticks;
}

// Force a reset event.
void animate_trigger_reset(void) {
	previous_animation_type = -1;
}

void animate(enum animation_event event) {
	
	static uint8_t angle;
	static int8_t angle_delta;
	static enum light selected_light;
	
	switch (animation_type) {
		case ANIMATION_TYPE_SIN_SLOW:
		case ANIMATION_TYPE_SIN_FAST:
		case ANIMATION_TYPE_CHASER:
			
			// Update each light.
			for (enum light l = LIGHT_A; l <= LIGHT_D; ++l) {
				
				// Grab the base sin value.
				uint8_t value = sin_uint8(angle += 64);
				
				// Do we need to "steepen" the value?
				// This makes the lights appear to flash rather than fade.
				switch (animation_type) {
					case ANIMATION_TYPE_CHASER:
						if (value >= 170) {
							value = (value - 170) * 3;
						} else {
							value = 0;
						}
						break;
					default:
						break;
				}
				light_set_level(l, value);
			}
			
			// Increment the angle (minus the 256 added above).
			angle = (angle + 1) - 256;
			
			// Schedule the next update based on the current animation type.
			switch (animation_type) {
				case ANIMATION_TYPE_SIN_SLOW: animate_set_timer(ANIMATION_TIMER_A, 350); break;
				case ANIMATION_TYPE_SIN_FAST: animate_set_timer(ANIMATION_TIMER_A, 80); break;
				case ANIMATION_TYPE_CHASER: animate_set_timer(ANIMATION_TIMER_A, 60); break;
				default: break;
			}
			break;
			
		case ANIMATION_TYPE_RANDOM_FADE:
		case ANIMATION_TYPE_RANDOM_FLASH:
			
			if (event == ANIMATION_EVENT_RESET) {
				enum light previous_selected_light = selected_light;
				do {
					selected_light = rand() % ((LIGHT_D - LIGHT_A) + 1);
				} while (selected_light == previous_selected_light);
				angle = 192;
				angle_delta = +1;
			}
			
			// Update each light.
			for (enum light l = LIGHT_A; l <= LIGHT_D; ++l) {
				
				// Grab the base sin value.
				uint8_t value = sin_uint8(angle);
				
				// If it's not the selected light, blank it out.
				if (l != selected_light) {
					value = 0;
				}
				
				light_set_level(l, value);
			}
			
			angle += angle_delta;
			
			if (angle == 64) {
				switch (animation_type) {
					case ANIMATION_TYPE_RANDOM_FADE: animate_set_timer(ANIMATION_TIMER_A, 60000); break;
					case ANIMATION_TYPE_RANDOM_FLASH: animate_set_timer(ANIMATION_TIMER_A, 6000); break;
					default: break;
				}
				angle_delta = -1;
			} else if (angle == 192) {
				animate_trigger_reset();
			} else {
				switch (animation_type) {
					case ANIMATION_TYPE_RANDOM_FADE: animate_set_timer(ANIMATION_TIMER_A, 500); break;
					case ANIMATION_TYPE_RANDOM_FLASH: animate_set_timer(ANIMATION_TIMER_A, 10); break;
					default: break;
				}
			}
			break;
			
		default:
			break;
	}
}