#include <stdint.h>

#include <avr/io.h>
#include <avr/interrupt.h>

#include "light.h"

// Initialise the lights. Interrupts should be enabled afterwards.
void light_init(void) {
	
	// Enable both timers for fast 8-bit PWM with no prescaler.
	TCCR0A |= _BV(WGM01) | _BV(WGM00);
	TCCR0B |= (0b001 << CS00);
	
	TCCR1A |= _BV(WGM12) | _BV(WGM10);
	TCCR1B |= (0b001 << CS10);
	
	// Generate interrupts
	TIMSK |= _BV(TOIE1) | _BV(TOIE0);
	
	// Configure the light I/O pins as outputs.
	LIGHT_A_DDR |= _BV(LIGHT_A_BIT);
	LIGHT_B_DDR |= _BV(LIGHT_B_BIT);
	LIGHT_C_DDR |= _BV(LIGHT_C_BIT);
	LIGHT_D_DDR |= _BV(LIGHT_D_BIT);
	
}

// Stores the "pending" timer OC register value for each of the lights.
// Rather than update the OC registers directly, they are updated at the end of a software PWM cycle.
// This avoids unsightly bright flickers when changing smoothly from one brightness to another.
uint8_t light_a_oc_pending = 255;
uint8_t light_b_oc_pending = 255;
uint8_t light_c_oc_pending = 255;
uint8_t light_d_oc_pending = 255;

// Sets the brightness of a single light.
void light_set_level(enum light light, uint8_t level) {
	level = 255 - level;
	switch (light) {
		case LIGHT_A: light_a_oc_pending = level; break;
		case LIGHT_B: light_b_oc_pending = level; break;
		case LIGHT_C: light_c_oc_pending = level; break;
		case LIGHT_D: light_d_oc_pending = level; break;
	}
}

// Update lights A and B in response to a timer interrupt.
void light_timer_update_ab(void) {
	// Perform software PWM.
	static volatile uint8_t software_pwm_0 = 0;
	if (software_pwm_0 == 0) {
		TCCR0A &= ~(_BV(COM0A1) | _BV(COM0A0) | _BV(COM0B1) | _BV(COM0B0));
		LIGHT_A_OC = light_a_oc_pending;
		LIGHT_B_OC = light_b_oc_pending;
	}
	if (LIGHT_A_OC == software_pwm_0) TCCR0A |= _BV(COM0A1) | _BV(COM0A0);
	if (LIGHT_B_OC == software_pwm_0) TCCR0A |= _BV(COM0B1) | _BV(COM0B0);
	++software_pwm_0;
}

// Update lights C and D in response to a timer interrupt.
void light_timer_update_cd(void) {
	// Perform software PWM.
	static volatile uint8_t software_pwm_1 = 0;
	if (software_pwm_1 == 0) {
		TCCR1A &= ~(_BV(COM1A1) | _BV(COM1A0) | _BV(COM1B1) | _BV(COM1B0));
		LIGHT_C_OC = light_c_oc_pending;
		LIGHT_D_OC = light_d_oc_pending;
	}
	if (LIGHT_C_OC == software_pwm_1) TCCR1A |= _BV(COM1A1) | _BV(COM1A0);
	if (LIGHT_D_OC == software_pwm_1) TCCR1A |= _BV(COM1B1) | _BV(COM1B0);
	++software_pwm_1;
}